//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "Unit1.h"
#include "math.h"
#include <stdlib.h>
#include <fstream.h>

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TForm1 *Form1;
//---------------------------------------------------------------------------
__fastcall TForm1::TForm1(TComponent* Owner)
    : TForm(Owner)
{
}
//---------------------------------------------------------------------------

const double PI = 3.1415926535;

struct Quat {
    double r;
    double i;
    double j;
    double k;
};

AnsiString quatToString(Quat *q) {
    return "Q[" +
        AnsiString(q->r) + "  " +
        AnsiString(q->i) + "  " +
        AnsiString(q->j) + "  " +
        AnsiString(q->k) + "]";
}

Quat *multiply(Quat *q1, Quat *q2) {
    Quat *r = new Quat;
    r->r = q1->r * q2->r - q1->i * q2->i - q1->j * q2->j - q1->k * q2->k;
    r->i = q1->r * q2->i + q1->i * q2->r - q1->j * q2->k + q1->k * q2->j;
    r->j = q1->r * q2->j + q1->i * q2->k + q1->j * q2->r - q1->k * q2->i;
    r->k = q1->r * q2->k - q1->i * q2->j + q1->j * q2->i + q1->k * q2->r;
    return r;
}

Quat *multiply(Quat *q, double scalar) {
    Quat *result = new Quat;
    result->r = scalar * q->r;
    result->i = scalar * q->i;
    result->j = scalar * q->j;
    result->k = scalar * q->k;
    return result;
}

Quat *add(Quat *q1, Quat *q2) {
    Quat *result = new Quat;
    result->r = q1->r + q2->r;
    result->i = q1->i + q2->i;
    result->j = q1->j + q2->j;
    result->k = q1->k + q2->k;
    return result;
}

Quat *normalize(Quat *q) {
    Quat *result = new Quat;

    double magnitude = sqrt(q->r * q->r + q->i * q->i + q->j * q->j + q->k * q->k);

    result->r = q->r / magnitude;
    result->i = q->i / magnitude;
    result->j = q->j / magnitude;
    result->k = q->k / magnitude;

    return result;
}


//---------------------------------------------------------------------------

struct Vector {
    double e1;
    double e2;
    double e3;
};

AnsiString vectorToString(Vector *v) {
    return "[" +
        AnsiString((int)(v->e1 * 1000 + 0.5) / 1000.0) + "  " +
        AnsiString((int)(v->e2 * 1000 + 0.5) / 1000.0) + "  " +
        AnsiString((int)(v->e3 * 1000 + 0.5) / 1000.0) + "]";
}

Vector *multiply(Vector *v, double scalar) {
    Vector *result = new Vector;
    result->e1 = v->e1 * scalar;
    result->e2 = v->e2 * scalar;
    result->e3 = v->e3 * scalar;
    return result;
}

double getDotProduct(Vector *v1, Vector *v2) {
    return v1->e1 * v2->e1 + v1->e2 *  v2->e2 + v1->e3 * v2->e3;
}

Vector *getCrossProduct(Vector *v1, Vector *v2) {
    Vector *result = new Vector;
    result->e1 = v1->e2 * v2->e3 - v2->e2 * v1->e3;
    result->e2 = v1->e3 * v2->e1 - v2->e3 * v1->e1;
    result->e3 = v1->e1 * v2->e2 - v2->e1 * v1->e2;
    return result;

}

Vector *add(Vector *v1, Vector *v2){
    Vector *result = new Vector;
    result->e1 = v1->e1 + v2->e1;
    result->e2 = v1->e2 + v2->e2;
    result->e3 = v1->e3 + v2->e3;
    return result;
}

Vector *subtract(Vector *v1, Vector *v2) {
    return add(v1, multiply(v2, -1.0));
}

Vector *getZeroVector() {
    Vector *result = new Vector;

    result->e1 = 0.0;
    result->e2 = 0.0;
    result->e3 = 0.0;

    return result;
}

//---------------------------------------------------------------------------

struct Matrix {
    double e11;
    double e12;
    double e13;
    double e21;
    double e22;
    double e23;
    double e31;
    double e32;
    double e33;
};


AnsiString matrixToString(Matrix *m) {
    return "M[" +
        AnsiString(m->e11) + "  " +
        AnsiString(m->e12) + "  " +
        AnsiString(m->e13) + "    " +
        AnsiString(m->e21) + "  " +
        AnsiString(m->e22) + "  " +
        AnsiString(m->e23) + "    " +
        AnsiString(m->e31) + "  " +
        AnsiString(m->e32) + "  " +
        AnsiString(m->e33) + "]";
}

Matrix *add(Matrix *m1, Matrix *m2) {
    Matrix *result = new Matrix();
    result->e11 = m1->e11 + m2->e11;
    result->e12 = m1->e12 + m2->e12;
    result->e13 = m1->e13 + m2->e13;
    result->e21 = m1->e21 + m2->e21;
    result->e22 = m1->e22 + m2->e22;
    result->e23 = m1->e23 + m2->e23;
    result->e31 = m1->e31 + m2->e31;
    result->e32 = m1->e32 + m2->e32;
    result->e33 = m1->e33 + m2->e33;
    return result;

}

Matrix *multiply(Matrix *m, double scalar) {
    Matrix *result = new Matrix;
    result->e11 = m->e11 * scalar;
    result->e12 = m->e12 * scalar;
    result->e13 = m->e13 * scalar;
    result->e21 = m->e21 * scalar;
    result->e22 = m->e22 * scalar;
    result->e23 = m->e23 * scalar;
    result->e31 = m->e31 * scalar;
    result->e32 = m->e32 * scalar;
    result->e33 = m->e33 * scalar;
    return result;
}

Matrix *getZeroMatrix() {
    Matrix *result = new Matrix();
    result->e11 = 0.0;
    result->e12 = 0.0;
    result->e13 = 0.0;
    result->e21 = 0.0;
    result->e22 = 0.0;
    result->e23 = 0.0;
    result->e31 = 0.0;
    result->e32 = 0.0;
    result->e33 = 0.0;
    return result;
}

Matrix *getIdentityMatrix() {
    Matrix *result = new Matrix();
    result->e11 = 1.0;
    result->e12 = 0.0;
    result->e13 = 0.0;
    result->e21 = 0.0;
    result->e22 = 1.0;
    result->e23 = 0.0;
    result->e31 = 0.0;
    result->e32 = 0.0;
    result->e33 = 1.0;
    return result;
}

Matrix *getOuterProduct(Vector *v1, Vector *v2) {
    Matrix *result = new Matrix();
    result->e11 = v1->e1 * v2->e1;
    result->e12 = v1->e1 * v2->e2;
    result->e13 = v1->e1 * v2->e3;
    result->e21 = v1->e2 * v2->e1;
    result->e22 = v1->e2 * v2->e2;
    result->e23 = v1->e2 * v2->e3;
    result->e31 = v1->e3 * v2->e1;
    result->e32 = v1->e3 * v2->e2;
    result->e33 = v1->e3 * v2->e3;
    return result;
}

Matrix *getInverseMatrix(Matrix *m) {
    Matrix *result = new Matrix();

    double a = m->e11;
    double b = m->e12;
    double c = m->e13;
    double d = m->e21;
    double e = m->e22;
    double f = m->e23;
    double g = m->e31;
    double h = m->e32;
    double k = m->e33;

    result->e11 = e * k - f * h;
    result->e12 = c * h - b * k;
    result->e13 = b * f - c * e;
    result->e21 = f * g - d * k;
    result->e22 = a * k - c * g;
    result->e23 = c * d - a * f;
    result->e31 = d * h - e * g;
    result->e32 = b * g - a * h;
    result->e33 = a * e - b * d;

    double z = a * (e * k - f * h) + b * (f * g - k * d) + c * (d * h - e * g);

    return multiply(result, 1.0 / z);
}

Matrix *getTransposed(Matrix *m) {
    Matrix *result = new Matrix;
    result->e11 = m->e11;
    result->e12 = m->e21;
    result->e13 = m->e31;
    result->e21 = m->e12;
    result->e22 = m->e22;
    result->e23 = m->e32;
    result->e31 = m->e13;
    result->e32 = m->e23;
    result->e33 = m->e33;
    return result;
}

Matrix *getRotMatrix(Quat *q) {
    Matrix *result = new Matrix;

    result->e11 = 1 -  2 * q->j * q->j - 2 * q->k * q->k;
    result->e12 =      2 * q->i * q->j - 2 * q->r * q->k;
    result->e13 =      2 * q->i * q->k + 2 * q->r * q->j;

    result->e21 =      2 * q->i * q->j + 2 * q->r * q->k;
    result->e22 = 1 -  2 * q->i * q->i - 2 * q->k * q->k;
    result->e23 =      2 * q->j * q->k - 2 * q->r * q->i;

    result->e31 =      2 * q->i * q->k - 2 * q->r * q->j;
    result->e32 =      2 * q->j * q->k + 2 * q->r * q->i;
    result->e33 = 1 -  2 * q->i * q->i - 2 * q->j * q->j;

    return result;
}

Matrix *multiply(Matrix *m1, Matrix *m2) {
    Matrix *result = new Matrix;
    result->e11 = m1->e11 * m2->e11 + m1->e12 * m2->e21 + m1->e13 * m2->e31;
    result->e12 = m1->e11 * m2->e12 + m1->e12 * m2->e22 + m1->e13 * m2->e32;
    result->e13 = m1->e11 * m2->e13 + m1->e12 * m2->e23 + m1->e13 * m2->e33;

    result->e21 = m1->e21 * m2->e11 + m1->e22 * m2->e21 + m1->e23 * m2->e31;
    result->e22 = m1->e21 * m2->e12 + m1->e22 * m2->e22 + m1->e23 * m2->e32;
    result->e23 = m1->e21 * m2->e13 + m1->e22 * m2->e23 + m1->e23 * m2->e33;

    result->e31 = m1->e31 * m2->e11 + m1->e32 * m2->e21 + m1->e33 * m2->e31;
    result->e32 = m1->e31 * m2->e12 + m1->e32 * m2->e22 + m1->e33 * m2->e32;
    result->e33 = m1->e31 * m2->e13 + m1->e32 * m2->e23 + m1->e33 * m2->e33;

    return result;
}

Vector *multiply(Matrix *m, Vector *v) {
    Vector *result = new Vector;

    result->e1 = m->e11 * v->e1 + m->e12 * v->e2 + m->e13 * v->e3;
    result->e2 = m->e21 * v->e1 + m->e22 * v->e2 + m->e23 * v->e3;
    result->e3 = m->e31 * v->e1 + m->e32 * v->e2 + m->e33 * v->e3;

    return result;
}

//---------------------------------------------------------------------------

double N = 8;
double DENSITY = 1.0;
const double UNIT = 4.0;

struct Particle {
    double radius;
    Vector *centerOfMass;
    double mass;
    Matrix *inertiaTensor;
};

void initParticle(Particle *p, Vector *c, double r) {
    p->radius = r;
    p->centerOfMass = c;
    p->mass = DENSITY * (PI * r * r);
    p->inertiaTensor = getZeroMatrix();
    double d = 2.0 * (p->mass) * (r * r) / 5.0;
    p->inertiaTensor->e11 = d;
    p->inertiaTensor->e22 = d;
    p->inertiaTensor->e33 = d;

}

AnsiString particleToString(Particle *p) {
    return
        "------PARTICLE------\r\n"
        "radius=" + AnsiString(p->radius) + "\r\n" +
        "centerOfMass=" + vectorToString(p->centerOfMass) + "\r\n" +
        "mass=" + AnsiString(p->mass) + "\r\n" +
        "inertiaTensor=" + matrixToString(p->inertiaTensor) + "\r\n";
}

void drawParticle(Particle *p, Vector *c, Matrix *rot) {

    Vector *v = multiply(rot, p->centerOfMass);
    double x1 = (c->e1 + v->e1 - p->radius) * UNIT;
    double y1 = Form1->Image1->Height - (c->e2 + v->e2 - p->radius) * UNIT;
    double x2 = x1 + (2 * p->radius) * UNIT;
    double y2 = y1 - (2 * p->radius) * UNIT;
    Form1->Image1->Canvas->Brush->Color = clBlue;
    Form1->Image1->Canvas->Ellipse(x1, y1, x2, y2);
}

//---------------------------------------------------------------------------
struct Body {
    Particle *p;    //array of N particles

    double mass;
    Matrix *inertiaTensorBody;
    Matrix *inertiaTensorBodyInv;

    Vector *centerOfMass;
    Quat *quat;
    Vector *linMomentum;
    Vector *angMomentum;

    Matrix *inertTensorInv;
    Matrix *rotMatrix;
    Vector *linVelocity;
    Vector *angVelocity;

    Vector *force;
    Vector *torque;
};

Body *b;

//----------------------------------------------------------------------------


init () {

    DENSITY = Form1->Edit1->Text.ToDouble();

    ifstream in;
    in.open("data.txt");
    in >> N;
    b->p = new Particle[N];

    for (int i = 0; i < N; i++) {
        Vector *v = new Vector;
        in >> v->e1;
        in >> v->e2;
        v->e3 = 0.0;
        initParticle(&(b->p[i]), v, 1.0);

    }
    in.close();

    b->mass = 0.0;
    for (int i = 0; i < N; i++) {
        b->mass += (&(b->p[i]))->mass;
    }

    b->centerOfMass = getZeroVector();

    for (int i = 0; i < N; i++) {
        b->centerOfMass = add(b->centerOfMass, multiply(b->p[i].centerOfMass, b->p[i].mass));
    }

    b->centerOfMass = multiply(b->centerOfMass, 1.0 / b->mass);

    for (int i = 0; i < N; i++) {
        b->p[i].centerOfMass = subtract(b->p[i].centerOfMass, b->centerOfMass);
    }

    b->inertiaTensorBody = getZeroMatrix();
    for (int i = 0; i < N; i++) {
        Vector *r = b->p[i].centerOfMass;
        double m = b->p[i].mass;
        Matrix *it = b->p[i].inertiaTensor;

        Matrix *m1;
        Matrix *m2;
        Matrix *m3;

        m1 = multiply(getIdentityMatrix(), getDotProduct(r, r));
        m2 = multiply(getOuterProduct(r, r), -1.0);
        m3 = multiply(add(m1, m2), m);
        b->inertiaTensorBody = add(b->inertiaTensorBody, add(it, m3));
    }

    b->inertiaTensorBodyInv = getInverseMatrix(b->inertiaTensorBody);

    b->quat = new Quat;
    b->quat->r = 1.0;
    b->quat->i = 0.0;
    b->quat->j = 0.0;
    b->quat->k = 0.0;

    b->linMomentum = getZeroVector();
    b->angMomentum = getZeroVector();

    b->force = getZeroVector();
    b->torque = getZeroVector();

    b->linVelocity = getZeroVector();
    b->angVelocity = getZeroVector();

    b->rotMatrix = getIdentityMatrix();
    b->inertTensorInv = getInverseMatrix(b->rotMatrix);

}

void step(double t) {

    b->centerOfMass = add(b->centerOfMass, multiply(b->linVelocity, t));

    if (
            (b->centerOfMass->e1 <= 0) ||
            (b->centerOfMass->e1 * UNIT >= Form1->Image1->Width) ||
            (b->centerOfMass->e2 <= 0) ||
            (b->centerOfMass->e2 * UNIT >= Form1->Image1->Height)) {

        b->linMomentum = multiply(b->linMomentum, -1.0);
        b->force = multiply(b->force, -1.0);

        }


    Quat *omega = new Quat;
    omega->r = 0;
    omega->i = b->angVelocity->e1;
    omega->j = b->angVelocity->e2;
    omega->k = b->angVelocity->e3;

    b->quat = add(b->quat, multiply(multiply(multiply(omega, b->quat), 0.5), t));

    b->linMomentum = add(b->linMomentum, multiply(b->force, t));
    b->angMomentum = add(b->angMomentum, multiply(b->torque, t));

    b->linVelocity = multiply(b->linMomentum, 1.0 / b->mass);
    b->rotMatrix = getRotMatrix(normalize(b->quat));

    b->inertTensorInv = multiply(b->rotMatrix, multiply(b->inertiaTensorBodyInv, getTransposed(b->rotMatrix)));
    b->angVelocity = multiply(b->inertTensorInv, b->angMomentum);

}


void draw(double clear = true) {

    if (clear) {
        Form1->Image1->Canvas->FillRect(Form1->Image1->Canvas->ClipRect);
    }

    for (int i = 0; i < N; i++) {
        drawParticle(&(b->p[i]), b->centerOfMass, b->rotMatrix);
    }

    Form1->Image1->Canvas->Brush->Color = clWhite;
    Form1->Image1->Canvas->FillRect(Rect(0, 0, 70, 70));

    Form1->Image1->Canvas->TextOutA
        (10, 10, "force = " + vectorToString(b->force));
    Form1->Image1->Canvas->TextOutA
        (10, 30, "linear velocity = " + vectorToString(b->linVelocity));
    Form1->Image1->Canvas->TextOutA
        (10, 50, "acceleration = " + vectorToString(multiply(b->force, 1.0 / b->mass)));
}



//---------------------------------------------------------------------------


void __fastcall TForm1::Button3Click(TObject *Sender)
{
    if (Timer1->Enabled) {
        init();
    }
    Timer1->Enabled = !Timer1->Enabled;
    draw();
    
}
//---------------------------------------------------------------------------


void __fastcall TForm1::Timer1Timer(TObject *Sender)
{
    step(0.2);
    draw();
}

int x1, x2, y1, y2;
int particleIndex;

//---------------------------------------------------------------------------

void __fastcall TForm1::FormCreate(TObject *Sender)
{
    DoubleBuffered = true;

    x1 = -1;
    x2 = -1;
    y1 = -1;
    y2 = -1;
    particleIndex = -1;

    b = new Body;
    init();
    draw();

}

void drawArrow(int x1, int y1, int x2, int y2) {

    double slopy , cosy , siny;
    double Par = 8.0;

    slopy = atan2(y1 - y2, x1 - x2);

    cosy = cos( slopy );
    siny = sin( slopy );

    Form1->Image1->Canvas->MoveTo(x1, y1);
    Form1->Image1->Canvas->LineTo(x2, y2);

    Form1->Image1->Canvas->MoveTo(x1, y1);
    Form1->Image1->Canvas->LineTo(
        x1 + int(-Par * cosy - (Par / 2.0 * siny)),
        y1 + int(-Par * siny + (Par / 2.0 * cosy)));

    Form1->Image1->Canvas->MoveTo(
        x1 + int(- Par * cosy + (Par / 2.0 * siny)),
        y1 - int(Par / 2.0 * cosy + Par * siny));

    Form1->Image1->Canvas->LineTo(x1, y1);
}

//---------------------------------------------------------------------------



int find(int X, int Y) {

    for (int i = 0; i < N; i++) {

        double dx = (b->p[i].centerOfMass->e1 + b->centerOfMass->e1) * UNIT - X;
        double dy =
            (b->p[i].centerOfMass->e2  + b->centerOfMass->e2) * UNIT -
            (Form1->Image1->Height - Y);
        double r = b->p[i].radius;

        if (sqrt(dx * dx + dy * dy) <= r * UNIT) {
            return i;
        }
    }
    return -1;

}

void __fastcall TForm1::Image1MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    particleIndex = find(X, Y);
    if (particleIndex != -1) {
        x1 = (b->p[particleIndex].centerOfMass->e1 + b->centerOfMass->e1) * UNIT;
        y1 = (Image1->Height -
            (b->p[particleIndex].centerOfMass->e2 + b->centerOfMass->e2) * UNIT);
    }

}
//---------------------------------------------------------------------------



void __fastcall TForm1::Image1MouseUp(TObject *Sender, TMouseButton Button,
      TShiftState Shift, int X, int Y)
{
    if (particleIndex == -1) {
        return;
    }
    

    x2 = X;
    y2 = Y;

    drawArrow(x2, y2, x1, y1);

    Vector *force = new Vector;

    force->e1 = x2 - x1;
    force->e2 = y1 - y2;
    force->e3 = 0.0;

    b->force = add(b->force, force);

    b->torque = add(b->torque, getCrossProduct(b->p[particleIndex].centerOfMass, force));

    draw(false);

    x1 = -1;
    x2 = -1;
    y1 = -1;
    y2 = -1;
    particleIndex = -1;


}
//---------------------------------------------------------------------------




void __fastcall TForm1::Edit1Change(TObject *Sender)
{
    Timer1->Enabled = false;
    init();
    draw();
}
//---------------------------------------------------------------------------

